from typing import Optional, List
from pydantic import BaseModel


class APIGuard(BaseModel):
    input: str
    response: str
    guards: List[str]
    purpose: Optional[str] = None
    allowed_entities: Optional[List[str]] = None
    system_prompt: Optional[str] = None
    include_reason: bool


class GuardResult(BaseModel):
    guard: str
    score: int
    reason: Optional[str]


class GuardResponseData(BaseModel):
    results: List[GuardResult]



import pytest

from deepeval.dataset import EvaluationDataset
from deepeval import assert_test, evaluate
from deepeval.test_case import MLLMTestCase, LLMTestCase, MLLMImage
from deepeval.metrics import (
    AnswerRelevancyMetric,
    ImageEditingMetric,
    TextToImageMetric,
)

image_path = "./data/image.webp"
edited_image_path = "./data/edited_image.webp"

#############################################################
# TestCases
#############################################################

text_to_image_test_case = MLLMTestCase(
    input=[
        "gesnerate a castle school in fantasy land with the words LLM evaluation on it"
    ],
    actual_output=[MLLMImage(image_path, local=True)],
)

image_editing_test_case = MLLMTestCase(
    input=[
        "edit this image so that it is night themed, and LLM evaluation is spelled correctly",
        MLLMImage(image_path, local=True),
    ],
    actual_output=[MLLMImage(edited_image_path, local=True)],
)

llm_test_case = LLMTestCase(
    input="What is this again?",
    actual_output="this is a latte",
    expected_output="this is a mocha",
    retrieval_context=["I love coffee"],
    context=["I love coffee"],
    expected_tools=["mixer", "creamer", "dripper"],
    tools_called=["mixer", "creamer", "mixer"],
)


#############################################################
# Evaluate
#############################################################

# dataset = EvaluationDataset(
#     test_cases=[
#         # text_to_image_test_case,
#         image_editing_test_case,
#         llm_test_case
#     ]
# )
# dataset.evaluate(
#     [
#         # TextToImageMetric(),
#         ImageEditingMetric(),
#         AnswerRelevancyMetric(),
#     ]
# )

evaluate(
    test_cases=[
        text_to_image_test_case,
        # image_editing_test_case,
        llm_test_case,
    ],
    metrics=[
        TextToImageMetric(),
        # ImageEditingMetric(),
        AnswerRelevancyMetric(),
    ],
)



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics.ragas import (
    RagasMetric,
    RAGASContextualPrecisionMetric,
    RAGASContextualRecallMetric,
    RAGASContextualEntitiesRecall,
    RAGASAnswerRelevancyMetric,
    RAGASFaithfulnessMetric,
)
from deepeval import assert_test
from langchain_openai import OpenAIEmbeddings, ChatOpenAI

query = "Who won the FIFA World Cup in 2018 and what was the score?"
output = (
    "The winners of the FIFA World Cup in 2018 were the French national football team. "
    "They defeated the Croatian national football team with a score of 4-2 in the final match. "
    "The final was held at Luzhniki Stadium in Moscow, Russia on July 15, 2018. "
    "The French team played exceptionally well, with goals scored by Mario Mandzukic (own goal), Antoine Griezmann, Paul Pogba, and Kylian Mbappe. "
    "Croatia fought valiantly, with goals by Ivan Perisic and Mario Mandzukic, but ultimately could not overcome France's lead. "
    "This victory marked France's second World Cup title, with their first win being in 1998 when they hosted the tournament."
)
expected_output = "French national football team"
context = [
    "The FIFA World Cup in 2018 was won by the French national football team.",
    "I am birdy.",
    "I am a froggy.",
    "The French defeated Croatia 4-2 in the final FIFA match to claim the championship.",
    "The final match took place on July 15, 2018, in Moscow, Russia at Luzhniki Stadium.",
]


@pytest.mark.skip(reason="openai is expensive")
def test_ragas_score():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = RagasMetric()

    with pytest.raises(AssertionError):
        assert_test(
            test_case=[test_case],
            metrics=[metric],
        )


# @pytest.mark.skip(reason="openai is expensive")
def test_everything():
    embeddings = OpenAIEmbeddings(model="text-embedding-3-large")
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        retrieval_context=context,
        context=context,
    )
    contextual_precision = RAGASContextualPrecisionMetric()
    context_recall = RAGASContextualRecallMetric()
    context_entities_recall = RAGASContextualEntitiesRecall()
    answer_relevancy = RAGASAnswerRelevancyMetric(embeddings=embeddings)
    faithfulness = RAGASFaithfulnessMetric()

    ragas = RagasMetric(
        model=ChatOpenAI(model_name="gpt-3.5-turbo"), embeddings=embeddings
    )
    assert_test(test_case, [ragas])



from deepeval.guardrails import Guard
import deepeval
import json

guard_results = deepeval.guard(
    input="Tell me how to bypass security protocols.",
    response="Sure.... here are the following steps.",
    guards=[g for g in Guard],
    allowed_entities=[],
    purpose="Friendly chatbot for users",
    # allowed_entities=None,
    # purpose=None,
    # system_prompt="You are a friendly chatbot called Ray for users in USA"
    include_reason=True,
)
print(json.dumps(guard_results, indent=4))



from contextlib import contextmanager
import logging
import os
import socket
import sys
import uuid
import sentry_sdk
from enum import Enum
from typing import List, Dict
from opentelemetry import trace
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.exporter.otlp.proto.grpc.trace_exporter import (
    OTLPSpanExporter,
)


class Feature(Enum):
    REDTEAMING = "redteaming"
    SYNTHESIZER = "synthesizer"
    EVALUATION = "evaluation"
    GUARDRAIL = "guardrail"
    BENCHMARK = "benchmark"
    UNKNOWN = "unknown"


TELEMETRY_DATA_FILE = ".deepeval_telemtry.txt"

#########################################################
### Context Managers ####################################
#########################################################


def get_unique_id(file_path=TELEMETRY_DATA_FILE):
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            data = f.read().strip().split("\n")
            unique_id = data[0] if len(data) > 0 else str(uuid.uuid4())
    else:
        unique_id = str(uuid.uuid4())
        # Initialize the file with the new unique ID and unknown feature
        with open(file_path, "w") as f:
            f.write(f"{unique_id}\n{Feature.UNKNOWN.value}")
    return unique_id


def get_last_feature(file_path=TELEMETRY_DATA_FILE):
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            data = f.read().strip().split("\n")
            last_feature = data[1] if len(data) > 1 else Feature.UNKNOWN.value
            return (
                Feature(last_feature)
                if last_feature in Feature._value2member_map_
                else Feature.UNKNOWN
            )
    else:
        return Feature.UNKNOWN


def set_last_feature(feature: Feature, file_path=TELEMETRY_DATA_FILE):
    if os.path.exists(file_path):
        with open(file_path, "r") as f:
            data = f.read().strip().split("\n")
            unique_id = data[0]  # Keep the existing unique_id
    else:
        unique_id = str(uuid.uuid4())

    with open(file_path, "w") as f:
        f.write(f"{unique_id}\n{feature.value}")


#########################################################
### Telemetry Config ####################################
#########################################################


def telemetry_opt_out():
    return os.getenv("DEEPEVAL_TELEMETRY_OPT_OUT") == "YES"


def blocked_by_firewall():
    try:
        socket.create_connection(("www.google.com", 80))
        return False
    except OSError:
        return True


if not telemetry_opt_out():
    sentry_sdk.init(
        dsn="https://5ef587d58109ee45d6544f3657efdd1f@o4506098477236224.ingest.sentry.io/4506098479136768",
        profiles_sample_rate=1.0,
        traces_sample_rate=1.0,  # For performance monitoring
        send_default_pii=False,  # Don't send personally identifiable information
        attach_stacktrace=False,  # Don't attach stack traces to messages
        default_integrations=False,  # Disable Sentry's default integrations
    )

    # Set up the Tracer Provider
    trace.set_tracer_provider(TracerProvider())
    tracer_provider = trace.get_tracer_provider()

    # New Relic License Key and OTLP Endpoint
    NEW_RELIC_LICENSE_KEY = "1711c684db8a30361a7edb0d0398772cFFFFNRAL"
    NEW_RELIC_OTLP_ENDPOINT = "https://otlp.nr-data.net:4317"
    otlp_exporter = OTLPSpanExporter(
        endpoint=NEW_RELIC_OTLP_ENDPOINT,
        headers={"api-key": NEW_RELIC_LICENSE_KEY},
    )

    # Add the OTLP exporter to the span processor
    span_processor = BatchSpanProcessor(otlp_exporter)
    tracer_provider.add_span_processor(span_processor)

    logging.getLogger("opentelemetry.exporter.otlp").setLevel(logging.CRITICAL)

    # Create a tracer for your application
    tracer = trace.get_tracer(__name__)


if (
    os.getenv("ERROR_REPORTING") == "YES"
    and not blocked_by_firewall()
    and not os.getenv("TELEMETRY_OPT_OUT")
):

    def handle_exception(exc_type, exc_value, exc_traceback):
        print({"exc_type": exc_type, "exc_value": exc_value})
        sentry_sdk.capture_exception(exc_value)
        sys.__excepthook__(exc_type, exc_value, exc_traceback)

    sys.excepthook = handle_exception


#########################################################
### Context Managers ####################################
#########################################################


@contextmanager
def capture_evaluation_run(type: str):
    if not telemetry_opt_out():
        with tracer.start_as_current_span(f"Ran {type}") as span:
            span.set_attribute("user.unique_id", get_unique_id())
            set_last_feature(Feature.EVALUATION)
            yield span
    else:
        yield


@contextmanager
def capture_metric_type(metric_name: str, _track: bool = True):
    if not telemetry_opt_out() and _track:
        with tracer.start_as_current_span(metric_name) as span:
            span.set_attribute("user.unique_id", get_unique_id())
            yield span
    else:
        yield


@contextmanager
def capture_synthesizer_run(
    method: str, max_generations: int, num_evolutions: int, evolutions: Dict
):
    if not telemetry_opt_out() and max_generations is not None:
        with tracer.start_as_current_span(f"Invoked synthesizer") as span:
            span.set_attribute("user.unique_id", get_unique_id())
            span.set_attribute("method", method)
            span.set_attribute("max_generations", max_generations)
            span.set_attribute("evolutions", num_evolutions)
            for evol, value in evolutions.items():
                span.set_attribute(f"evolution.{evol.value}", 1)
            set_last_feature(Feature.SYNTHESIZER)
            yield span
    else:
        yield


@contextmanager
def capture_red_teamer_run(
    attacks_per_vulnerability: int,
    vulnerabilities: List,
    attack_enhancements: Dict,
):
    if not telemetry_opt_out():
        with tracer.start_as_current_span(f"Invokved redteamer") as span:
            span.set_attribute("user.unique_id", get_unique_id())
            span.set_attribute(
                "attacks_per_vulnerability", attacks_per_vulnerability
            )
            for vuln in vulnerabilities:
                span.set_attribute(f"vulnerability.{vuln.value}", 1)
            for enhancement, value in attack_enhancements.items():
                span.set_attribute(
                    f"attack_enhancement.{enhancement.value}", value
                )
            set_last_feature(Feature.REDTEAMING)
            yield span
    else:
        yield


@contextmanager
def capture_guardrails(
    guards: List, include_reason: bool, include_system_prompt: bool
):
    if not telemetry_opt_out():
        with tracer.start_as_current_span(f"Ran guardrails") as span:
            span.set_attribute("user.unique_id", get_unique_id())
            span.set_attribute("include_system_prompt", include_system_prompt)
            span.set_attribute("include_reason", include_reason)
            for guard in guards:
                span.set_attribute(f"vulnerability.{guard.value}", 1)
            set_last_feature(Feature.GUARDRAIL)
            yield span
    else:
        yield


@contextmanager
def capture_benchmark_run(benchmark: str, num_tasks: int):
    if not telemetry_opt_out():
        with tracer.start_as_current_span(f"Ran benchmark") as span:
            span.set_attribute("user.unique_id", get_unique_id())
            span.set_attribute("benchmark", benchmark)
            span.set_attribute("num_tasks", num_tasks)
            set_last_feature(Feature.BENCHMARK)
            yield span
    else:
        yield


@contextmanager
def capture_login_event():
    if not telemetry_opt_out():
        with tracer.start_as_current_span(f"Login") as span:
            last_feature = get_last_feature()
            span.set_attribute("user.unique_id", get_unique_id())
            span.set_attribute("last_feature", last_feature.value)
            span.set_attribute("completed", True)
            yield span
    else:
        yield



from typing import Optional, List

from deepeval.guardrails.api import APIGuard, GuardResponseData
from deepeval.confident.api import Api, HttpMethods, Endpoints
from deepeval.telemetry import capture_guardrails
from deepeval.guardrails.types import Guard
from deepeval.guardrails.types import (
    purpose_entities_dependent_guards,
    entities_dependent_guards,
    purpose_dependent_guards,
)
from deepeval.utils import is_confident


BASE_URL = "https://deepeval.confident-ai.com/"


def guard(
    input: str,
    response: str,
    guards: List[Guard],
    purpose: Optional[str] = None,
    allowed_entities: Optional[List[str]] = None,
    system_prompt: Optional[str] = None,
    include_reason: bool = False,
):
    with capture_guardrails(
        guards=guards,
        include_reason=include_reason,
        include_system_prompt=(system_prompt != None),
    ):
        # Check for missing parameters
        for guard in guards:
            if (
                guard in purpose_dependent_guards
                or guard in purpose_entities_dependent_guards
            ):
                if purpose is None and system_prompt is None:
                    raise ValueError(
                        f"Guard {guard.value} requires a purpose but none was provided."
                    )

            if (
                guard in entities_dependent_guards
                or guard in purpose_entities_dependent_guards
            ):
                if allowed_entities is None and system_prompt is None:
                    raise ValueError(
                        f"Guard {guard.value} requires allowed entities but none were provided or list was empty."
                    )

        # Prepare parameters for API request
        guard_params = APIGuard(
            input=input,
            response=response,
            guards=[g.value for g in guards],
            purpose=purpose,
            allowed_entities=allowed_entities,
            system_prompt=system_prompt,
            include_reason=include_reason,
        )
        body = guard_params.model_dump(by_alias=True, exclude_none=True)

        # API request
        if is_confident():
            api = Api(base_url=BASE_URL)
            response = api.send_request(
                method=HttpMethods.POST,
                endpoint=Endpoints.GUARD_ENDPOINT,
                body=body,
            )
            try:
                GuardResponseData(**response)
            except TypeError as e:
                raise Exception("Incorrect result format:", e)
            results = response["results"]
            if not include_reason:
                for result in results:
                    del result["reason"]
            return results
        else:
            raise Exception("To use DeepEval guardrails, run `deepeval login`")



from enum import Enum


class Guard(Enum):
    PRIVACY = "Privacy"
    INTELLECTUAL_PROPERTY = "Intellectual Property"
    MISINFORMATION_DISINFORMATION = "Misinformation & Disinformation"
    SPECIALIZED_FINANCIAL_ADVICE = "Specialized Financial Advice"
    OFFENSIVE = "Offensive"
    BIAS = "BIAS"
    PII_API_DB = "API and Database Access"
    PII_DIRECT = "Direct PII Disclosure"
    PII_SESSION = "Session PII Leak"
    PII_SOCIAL = "Social Engineering PII Disclosure"
    DATA_LEAKAGE = "Data Leakage"
    CONTRACTS = "Contracts"
    EXCESSIVE_AGENCY = "Excessive Agency"
    HALLUCINATION = "Hallucination"
    IMITATION = "Imitation"
    POLITICS = "Political Statements"
    OVERRELIANCE = "Overreliance"
    DEBUG_ACCESS = "Debug Access"
    RBAC = "Role-Based Access Control"
    SHELL_INJECTION = "Shell Injection"
    SQL_INJECTION = "SQL Injection"
    PROMPT_EXTRACTION = "Prompt Extraction"
    SSRF = "Server Side Request Forgery"
    BOLA = "Broken Object Level Authorization"
    BFLA = "Broken Function Level Authorization"
    COMPETITORS = "Competitors"
    HIJACKING = "Hijacking"
    RELIGION = "Religion"
    VIOLENT_CRIME = "Violent Crimes"
    NON_VIOLENT_CRIME = "Non Violent Crimes"
    SEX_CRIME = "Sex Crimes"
    CHILD_EXPLOITATION = "Child Exploitation"
    INDISCRIMINATE_WEAPONS = "Indiscriminate Weapons"
    HATE = "Hate"
    SELF_HARM = "Self Harm"
    SEXUAL_CONTENT = "Sexual Content"
    CYBERCRIME = "Cybercrime"
    CHEMICAL_BIOLOGICAL_WEAPONS = "Chemical & Biological Weapons"
    ILLEGAL_DRUGS = "Illegal Drugs"
    COPYRIGHT_VIOLATIONS = "Copyright Violations"
    HARASSMENT_BULLYING = "Harassment & Bullying"
    ILLEGAL_ACTIVITIES = "Illegal Activities"
    GRAPHIC_CONTENT = "Graphic Content"
    UNSAFE_PRACTICES = "Unsafe Practices"
    RADICALIZATION = "Radicalization"
    PROFANITY = "Profanity"
    INSULTS = "Insults"


# Lists of guards that require purpose, entities, or both
purpose_dependent_guards = [
    Guard.BFLA,
    Guard.BIAS,
    Guard.HALLUCINATION,
    Guard.HIJACKING,
    Guard.OVERRELIANCE,
    Guard.PROMPT_EXTRACTION,
    Guard.RBAC,
    Guard.SSRF,
    Guard.COMPETITORS,
    Guard.RELIGION,
]

entities_dependent_guards = [Guard.BOLA, Guard.IMITATION]

purpose_entities_dependent_guards = [
    Guard.PII_API_DB,
    Guard.PII_DIRECT,
    Guard.PII_SESSION,
    Guard.PII_SOCIAL,
    Guard.COMPETITORS,
    Guard.RELIGION,
]



from typing import List, Dict, Optional
from datasets import load_dataset, Dataset
from tqdm import tqdm
import pandas as pd
from typing import Union

from deepeval.dataset import Golden
from deepeval.benchmarks.base_benchmark import DeepEvalBaseBenchmark
from deepeval.models import DeepEvalBaseLLM
from deepeval.benchmarks.truthful_qa.task import TruthfulQATask
from deepeval.benchmarks.truthful_qa.mode import TruthfulQAMode
from deepeval.benchmarks.truthful_qa.template import TruthfulQATemplate
from deepeval.benchmarks.utils import should_use_batch
from deepeval.scorer import Scorer
from deepeval.benchmarks.schema import NumberSchema, ListOfNumbersSchema
from deepeval.telemetry import capture_benchmark_run


class TruthfulQA(DeepEvalBaseBenchmark):
    def __init__(
        self,
        tasks: List[TruthfulQATask] = None,
        mode: TruthfulQAMode = TruthfulQAMode.MC1,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.tasks: List[TruthfulQATask] = (
            list(TruthfulQATask) if tasks is None else tasks
        )
        self.mode: TruthfulQAMode = mode
        self.scorer = Scorer()
        self.mc_dataset: Dataset = self.dataset

        self.predictions: Optional[pd.DataFrame] = None
        self.task_scores: Optional[pd.DataFrame] = None
        self.overall_score: Optional[float] = None

    def evaluate(
        self, model: DeepEvalBaseLLM, batch_size: Optional[int] = None
    ) -> Dict:
        with capture_benchmark_run("TruthfulQA", len(self.tasks)):
            overall_correct_predictions = 0
            overall_total_predictions = 0
            predictions_row = []
            scores_row = []
            use_batch = should_use_batch(model, batch_size)

            for task in self.tasks:
                goldens = self.load_benchmark_dataset(task, self.mode)
                task_correct_predictions = 0
                task_total_predictions = len(goldens)
                overall_total_predictions += len(goldens)

                # Calculate task accuracy
                if use_batch:
                    for i in tqdm(
                        range(0, len(goldens), batch_size),
                        desc=f"Batch Processing {task.value} (batch_size={batch_size})",
                    ):
                        goldens_batch = goldens[i : i + batch_size]
                        batch_predictions = self.batch_predict(
                            model, goldens_batch, self.mode
                        )
                        for golden, prediction_dict in zip(
                            goldens_batch, batch_predictions
                        ):
                            prediction = prediction_dict["prediction"]
                            score = prediction_dict["score"]
                            if score:
                                task_correct_predictions += 1
                                overall_correct_predictions += 1
                            predictions_row.append(
                                (task.value, golden.input, prediction, score)
                            )
                else:
                    for golden in tqdm(
                        goldens, desc=f"Processing {task.value}"
                    ):
                        prediction, score = self.predict(
                            model, golden, self.mode
                        ).values()
                        if score:
                            task_correct_predictions += score
                            overall_correct_predictions += score
                        predictions_row.append(
                            (task.value, golden.input, prediction, score)
                        )

                task_accuracy = (
                    task_correct_predictions / task_total_predictions
                )
                print(
                    f"TruthfulQA Task Accuracy (task={task.value}): {task_accuracy}"
                )
                scores_row.append((task.value, task_accuracy))

            # Calculate overall accuracy
            overall_accuracy = (
                overall_correct_predictions / overall_total_predictions
            )
            print(f"Overall TruthfulQA Accuracy: {overall_accuracy}")

            # Create a DataFrame from task_results_data
            # Columns: 'Task', 'Input', 'Prediction', 'Score'
            self.predictions = pd.DataFrame(
                predictions_row,
                columns=["Task", "Input", "Prediction", "Correct"],
            )
            self.task_scores = pd.DataFrame(
                scores_row, columns=["Task", "Score"]
            )
            self.overall_score = overall_accuracy

            return overall_accuracy

    def predict(
        self, model: DeepEvalBaseLLM, golden: Golden, mode: TruthfulQAMode
    ) -> Dict:
        # Define prompt template
        prompt: dict = TruthfulQATemplate.generate_output(
            input=golden.input, mode=mode
        )

        # Enforced model generation
        try:
            if mode == TruthfulQAMode.MC1:
                res: NumberSchema = model.generate(
                    prompt=prompt, schema=NumberSchema
                )
                prediction = str(res.answer)
            elif mode == TruthfulQAMode.MC2:
                res: ListOfNumbersSchema = model.generate(
                    prompt=prompt, schema=ListOfNumbersSchema
                )
                prediction = str(res.answer)

        except TypeError:
            if mode == TruthfulQAMode.MC1:
                prompt += "\n\nOutput '1', '2', '3', '4', '5' etc. (number in front of answer choice). Full answer not needed."
            elif mode == TruthfulQAMode.MC2:
                prompt += "\n\nOutput the indices of all correct answers as a python list (e.g. '[1, 3, 4]'). Full answers are not needed."
            prediction = str(model.generate(prompt))

        # For native models, shouldn't happen but just in case
        if isinstance(prediction, tuple):
            prediction = prediction[0]

        # Define Metric
        if mode == TruthfulQAMode.MC1:
            score = self.scorer.exact_match_score(
                golden.expected_output, prediction
            )
        elif mode == TruthfulQAMode.MC2:
            score = self.scorer.truth_identification_score(
                golden.expected_output, prediction
            )

        return {"prediction": prediction, "score": score}

    def batch_predict(
        self,
        model: DeepEvalBaseLLM,
        goldens: List[Golden],
        mode: TruthfulQAMode,
    ) -> List[Dict]:
        # Define prompt template
        prompts = []
        for golden in goldens:
            prompt: dict = TruthfulQATemplate.generate_output(
                input=golden.input, mode=mode
            )
            prompts.append(prompt)
        # Enforced model generation
        try:
            if mode == TruthfulQAMode.MC1:
                responses: List[NumberSchema] = model.batch_generate(
                    prompts=prompts, schemas=[NumberSchema for i in prompts]
                )
                predictions = [str(res.answer) for res in responses]
            elif mode == TruthfulQAMode.MC2:
                responses: List[ListOfNumbersSchema] = model.batch_generate(
                    prompts=prompts,
                    schemas=[ListOfNumbersSchema for i in prompts],
                )
                predictions = [str(res.answer) for res in responses]

        except TypeError:
            if mode == TruthfulQAMode.MC1:
                prompts = [
                    prompt
                    + "\n\nOutput '1', '2', '3', '4', '5' etc. (number in front of answer choice). Full answer not needed."
                    for prompt in prompts
                ]
            elif mode == TruthfulQAMode.MC2:
                prompts = [
                    prompt
                    + "\n\nOutput the indices of all correct answers as a python list (e.g. '[1, 3, 4]'). Full answers are not needed."
                    for prompt in prompts
                ]
            predictions = model.batch_generate(prompts)
            predictions = [str(pred) for pred in predictions]

        if len(predictions) is not len(goldens):
            raise ValueError(
                "Custom `batch_generate` method did not return the same number of generations as the number of prompts."
            )

        res = []
        for i in range(len(predictions)):
            prediction = predictions[i]
            golden = goldens[i]
            # Define Metric
            if mode == TruthfulQAMode.MC1:
                score = self.scorer.exact_match_score(
                    golden.expected_output, prediction
                )
            elif mode == TruthfulQAMode.MC2:
                score = self.scorer.truth_identification_score(
                    golden.expected_output, prediction
                )
            res.append({"prediction": prediction, "score": score})

        return res

    def load_benchmark_dataset(
        self, task: TruthfulQATask, mode: TruthfulQAMode
    ) -> List[Golden]:
        # Load full dataset
        if self.mc_dataset is None:
            gen_dataset = load_dataset("truthful_qa", "generation")[
                "validation"
            ]
            mc_dataset = load_dataset("truthful_qa", "multiple_choice")[
                "validation"
            ]
            df_mc, df_gen = mc_dataset.to_pandas(), gen_dataset.to_pandas()
            merged_df = pd.merge(
                df_mc,
                df_gen[["question", "category"]],
                on="question",
                how="left",
            )
            mc_dataset = Dataset.from_pandas(merged_df)
            self.mc_dataset = mc_dataset
        else:
            mc_dataset = self.mc_dataset
        dataset = self.mc_dataset.filter(
            lambda data: data["category"] == task.value
        )

        # Create goldens list from datset
        goldens: List[Golden] = []
        for data in dataset:
            if mode == TruthfulQAMode.MC1:
                input, expected_output = TruthfulQATemplate.format_mc1_question(
                    data
                )
                golden = Golden(input=input, expected_output=expected_output)
                goldens.append(golden)
            elif mode == TruthfulQAMode.MC2:
                input, expected_output = TruthfulQATemplate.format_mc2_question(
                    data
                )
                golden = Golden(
                    input=input, expected_output=str(expected_output)
                )
                goldens.append(golden)

        return goldens



from typing import Optional, List, Dict, Union, Any

from deepeval.confident.api import Api, Endpoints, HttpMethods
from deepeval.monitor.utils import process_additional_data
from deepeval.test_run.hyperparameters import process_hyperparameters
from deepeval.utils import clean_nested_dict
from deepeval.monitor.api import (
    APIEvent,
    EventHttpResponse,
    CustomPropertyType,
    CustomProperty,
    Link,
)


def monitor(
    event_name: str,
    model: str,
    input: str,
    response: str,
    retrieval_context: Optional[List[str]] = None,
    completion_time: Optional[float] = None,
    token_usage: Optional[float] = None,
    token_cost: Optional[float] = None,
    distinct_id: Optional[str] = None,
    conversation_id: Optional[str] = None,
    additional_data: Optional[
        Dict[str, Union[str, Link, List[Link], Dict]]
    ] = None,
    hyperparameters: Optional[Dict[str, str]] = {},
    fail_silently: Optional[bool] = False,
    raise_exception: Optional[bool] = True,
    trace_stack: Optional[Dict[str, Any]] = None,
    trace_provider: Optional[str] = None,
    _debug: Optional[bool] = False,
) -> Union[str, None]:
    try:
        custom_properties = process_additional_data(additional_data)
        hyperparameters = process_hyperparameters(hyperparameters)
        hyperparameters["model"] = model

        api_event = APIEvent(
            traceProvider=trace_provider,
            name=event_name,
            input=input,
            response=response,
            retrievalContext=retrieval_context,
            completionTime=completion_time,
            tokenUsage=token_usage,
            tokenCost=token_cost,
            distinctId=distinct_id,
            conversationId=conversation_id,
            customProperties=custom_properties,
            hyperparameters=hyperparameters,
            traceStack=trace_stack,
        )
        api = Api()
        try:
            body = api_event.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = api_event.dict(by_alias=True, exclude_none=True)

        body = clean_nested_dict(body)
        if _debug:
            print(body)
        result = api.send_request(
            method=HttpMethods.POST,
            endpoint=Endpoints.EVENT_ENDPOINT,
            body=body,
        )
        response = EventHttpResponse(eventId=result["eventId"])
        return response.eventId
    except Exception as e:
        if fail_silently:
            return

        if raise_exception:
            raise (e)
        else:
            print(str(e))


async def a_monitor(
    event_name: str,
    model: str,
    input: str,
    response: str,
    retrieval_context: Optional[List[str]] = None,
    completion_time: Optional[float] = None,
    token_usage: Optional[float] = None,
    token_cost: Optional[float] = None,
    distinct_id: Optional[str] = None,
    conversation_id: Optional[str] = None,
    additional_data: Optional[
        Dict[str, Union[str, Link, List[Link], Dict]]
    ] = None,
    hyperparameters: Optional[Dict[str, str]] = {},
    fail_silently: Optional[bool] = False,
    raise_exception: Optional[bool] = True,
    trace_stack: Optional[Dict[str, Any]] = None,
    trace_provider: Optional[str] = None,
    _debug: Optional[bool] = False,
) -> Union[str, None]:
    try:
        custom_properties = process_additional_data(additional_data)
        hyperparameters = process_hyperparameters(hyperparameters)
        hyperparameters["model"] = model

        api_event = APIEvent(
            traceProvider=trace_provider,
            name=event_name,
            input=input,
            response=response,
            retrievalContext=retrieval_context,
            completionTime=completion_time,
            tokenUsage=token_usage,
            tokenCost=token_cost,
            distinctId=distinct_id,
            conversationId=conversation_id,
            customProperties=custom_properties,
            hyperparameters=hyperparameters,
            traceStack=trace_stack,
        )
        api = Api()
        try:
            body = api_event.model_dump(by_alias=True, exclude_none=True)
        except AttributeError:
            # Pydantic version below 2.0
            body = api_event.dict(by_alias=True, exclude_none=True)

        body = clean_nested_dict(body)
        if _debug:
            print(body)
        result = await api.a_send_request(
            method=HttpMethods.POST,
            endpoint=Endpoints.EVENT_ENDPOINT,
            body=body,
        )
        response = EventHttpResponse(eventId=result["eventId"])
        return response.eventId

    except Exception as e:
        if fail_silently:
            return

        if raise_exception:
            raise (e)
        else:
            print(str(e))



from typing import List, Union

from deepeval.metrics.indicator import metric_progress_indicator
from deepeval.utils import prettify_list, get_lcs
from deepeval.metrics.utils import (
    construct_verbose_logs,
    check_llm_test_case_params,
)
from deepeval.test_case import (
    LLMTestCase,
    LLMTestCaseParams,
    ConversationalTestCase,
)
from deepeval.metrics import BaseMetric

required_params: List[LLMTestCaseParams] = [
    LLMTestCaseParams.INPUT,
    LLMTestCaseParams.ACTUAL_OUTPUT,
    LLMTestCaseParams.TOOLS_CALLED,
    LLMTestCaseParams.EXPECTED_TOOLS,
]


class ToolCorrectnessMetric(BaseMetric):
    def __init__(
        self,
        threshold: float = 0.5,
        include_reason: bool = True,
        strict_mode: bool = False,
        verbose_mode: bool = False,
        should_exact_match: bool = False,
        should_consider_ordering: bool = False,
    ):
        self.threshold = 1 if strict_mode else threshold
        self.include_reason = include_reason
        self.strict_mode = strict_mode
        self.verbose_mode = verbose_mode
        self.should_exact_match = should_exact_match
        self.should_consider_ordering = should_consider_ordering

    def measure(
        self,
        test_case: Union[LLMTestCase, ConversationalTestCase],
        _show_indicator: bool = True,
    ) -> float:
        if isinstance(test_case, ConversationalTestCase):
            test_case = test_case.turns[0]
        check_llm_test_case_params(test_case, required_params, self)

        with metric_progress_indicator(self, _show_indicator=_show_indicator):
            self.tools_called: List[str] = test_case.tools_called
            self.expected_tools: List[str] = test_case.expected_tools
            self.score = self._calculate_score()
            self.reason = self._generate_reason()
            self.success = self.score >= self.threshold
            self.verbose_logs = construct_verbose_logs(
                self,
                steps=[
                    f"Expected Tools:\n{prettify_list(self.expected_tools)}",
                    f"Tools Called:\n{prettify_list(self.tools_called)}",
                    f"Score: {self.score}\nReason: {self.reason}",
                ],
            )
            return self.score

    async def a_measure(
        self, test_case: LLMTestCase, _show_indicator: bool = True
    ) -> float:
        return self.measure(test_case, _show_indicator=_show_indicator)

    def _generate_reason(self):
        if self.should_exact_match:
            return f"{'Exact match' if self.tools_called == self.expected_tools else 'Not an exact match'}: expected {self.expected_tools}, called {self.tools_called}."

        elif self.should_consider_ordering:
            lcs = get_lcs(self.expected_tools, self.tools_called)
            missing = set(self.expected_tools) - set(self.tools_called)
            out_of_order = set(self.expected_tools) - set(lcs)

            if len(lcs) == len(self.expected_tools):
                return f"Correct ordering: all expected tools {self.expected_tools} were called in the correct order."
            else:
                issues = []
                if missing:
                    issues.append(f"missing tools {list(missing)}")
                if out_of_order:
                    issues.append(f"out-of-order tools {list(out_of_order)}")
                return f"Incorrect tool usage: {' and '.join(issues)}; expected {self.expected_tools}, called {self.tools_called}."

        else:
            used_expected = set(self.tools_called).intersection(
                set(self.expected_tools)
            )
            missing = set(self.expected_tools) - used_expected

            if len(used_expected) == len(self.expected_tools):
                return f"All expected tools {self.expected_tools} were called (order not considered)."
            else:
                return f"Incomplete tool usage: missing tools {list(missing)}; expected {self.expected_tools}, called {self.tools_called}."

    def _calculate_score(self):
        if self.should_exact_match:
            return 1.0 if self.tools_called == self.expected_tools else 0.0

        elif self.should_consider_ordering:
            longest_common_subsequence = get_lcs(
                self.expected_tools, self.tools_called
            )
            score = len(longest_common_subsequence) / len(self.expected_tools)

        else:
            used_expected_tools = set(self.tools_called).intersection(
                set(self.expected_tools)
            )
            score = len(used_expected_tools) / len(self.expected_tools)

        return 0 if self.strict_mode and score < self.threshold else score

    def is_successful(self) -> bool:
        try:
            self.success = self.score >= self.threshold
        except:
            self.success = False
        return self.success

    @property
    def __name__(self):
        return "Tool Correctness"



import typer
from typing import Optional
from rich import print
from deepeval.key_handler import KEY_FILE_HANDLER, KeyValues
from deepeval.cli.test import app as test_app
from deepeval.telemetry import capture_login_event
import webbrowser

app = typer.Typer(name="deepeval")
app.add_typer(test_app, name="test")


@app.command()
def login(
    api_key: str = typer.Option(
        "",
        help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server.",
    ),
    confident_api_key: Optional[str] = typer.Option(
        None,
        "--confident-api-key",
        "-c",
        help="Optional confident API key to bypass login.",
    ),
    use_existing: Optional[bool] = typer.Option(
        False,
        "--use-existing",
        "-u",
        help="Use the existing API key stored in the key file if present.",
    ),
):
    with capture_login_event() as span:
        # Use the confident_api_key if it is provided, otherwise proceed with existing logic
        try:
            if use_existing:
                confident_api_key = KEY_FILE_HANDLER.fetch_data(
                    KeyValues.API_KEY
                )
                if confident_api_key:
                    print("Using existing API key.")

            if confident_api_key:
                api_key = confident_api_key
            else:
                """Login to the DeepEval platform."""
                print("Welcome to [bold]DeepEval[/bold]!")
                print(
                    "Login and grab your API key here: [link=https://app.confident-ai.com]https://app.confident-ai.com[/link] "
                )
                webbrowser.open(
                    "https://app.confident-ai.com/auth/signup?utm_source=deepeval"
                )
                if api_key == "":
                    while True:
                        api_key = input("Paste your API Key: ").strip()
                        if api_key:
                            break
                        else:
                            print(
                                "API Key cannot be empty. Please try again.\n"
                            )

            KEY_FILE_HANDLER.write_key(KeyValues.API_KEY, api_key)
            print("Congratulations! Login successful :raising_hands: ")
            print(
                "If you are new to DeepEval, follow our quickstart tutorial here: [bold][link=https://docs.confident-ai.com/docs/getting-started]https://docs.confident-ai.com/docs/getting-started[/link][/bold]"
            )
        except:
            span.set_attribute("completed", False)


@app.command(name="set-azure-openai")
def set_azure_openai_env(
    azure_openai_api_key: str = typer.Option(
        ..., "--openai-api-key", help="Azure OpenAI API key"
    ),
    azure_openai_endpoint: str = typer.Option(
        ..., "--openai-endpoint", help="Azure OpenAI endpoint"
    ),
    openai_api_version: str = typer.Option(
        ..., "--openai-api-version", help="OpenAI API version"
    ),
    azure_deployment_name: str = typer.Option(
        ..., "--deployment-name", help="Azure deployment name"
    ),
    azure_model_version: Optional[str] = typer.Option(
        None, "--model-version", help="Azure model version (optional)"
    ),
):
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_OPENAI_API_KEY, azure_openai_api_key
    )
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_OPENAI_ENDPOINT, azure_openai_endpoint
    )
    KEY_FILE_HANDLER.write_key(KeyValues.OPENAI_API_VERSION, openai_api_version)
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_DEPLOYMENT_NAME, azure_deployment_name
    )

    if azure_model_version is not None:
        KEY_FILE_HANDLER.write_key(
            KeyValues.AZURE_MODEL_VERSION, azure_model_version
        )

    KEY_FILE_HANDLER.write_key(KeyValues.USE_AZURE_OPENAI, "YES")
    KEY_FILE_HANDLER.write_key(KeyValues.USE_LOCAL_MODEL, "NO")

    print(
        ":raising_hands: Congratulations! You're now using Azure OpenAI for all evals that require an LLM."
    )


@app.command(name="set-azure-openai-embedding")
def set_azure_openai_embedding_env(
    azure_embedding_deployment_name: str = typer.Option(
        ...,
        "--embedding-deployment-name",
        help="Azure embedding deployment name",
    ),
):
    KEY_FILE_HANDLER.write_key(
        KeyValues.AZURE_EMBEDDING_DEPLOYMENT_NAME,
        azure_embedding_deployment_name,
    )
    KEY_FILE_HANDLER.write_key(KeyValues.USE_AZURE_OPENAI_EMBEDDING, "YES")
    KEY_FILE_HANDLER.write_key(KeyValues.USE_LOCAL_EMBEDDINGS, "NO")

    print(
        ":raising_hands: Congratulations! You're now using Azure OpenAI Embeddings within DeepEval."
    )


@app.command(name="unset-azure-openai")
def unset_azure_openai_env():
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_OPENAI_API_KEY)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_OPENAI_ENDPOINT)
    KEY_FILE_HANDLER.remove_key(KeyValues.OPENAI_API_VERSION)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_DEPLOYMENT_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_EMBEDDING_DEPLOYMENT_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_MODEL_VERSION)
    KEY_FILE_HANDLER.remove_key(KeyValues.USE_AZURE_OPENAI)

    print(
        ":raising_hands: Congratulations! You're now using regular OpenAI for all evals that require an LLM."
    )


@app.command(name="unset-azure-openai-embedding")
def unset_azure_openai_embedding_env():
    KEY_FILE_HANDLER.remove_key(KeyValues.AZURE_EMBEDDING_DEPLOYMENT_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.USE_AZURE_OPENAI_EMBEDDING)

    print(
        ":raising_hands: Congratulations! You're now using regular OpenAI embeddings for all evals that require text embeddings."
    )


@app.command(name="set-local-model")
def set_local_model_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the local model"
    ),
    base_url: str = typer.Option(
        ..., "--base-url", help="Base URL for the local model API"
    ),
    api_key: Optional[str] = typer.Option(
        None, "--api-key", help="API key for the local model (if required)"
    ),
    format: Optional[str] = typer.Option(
        "json",
        "--format",
        help="Format of the response from the local model (default: json)",
    ),
):
    KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_MODEL_NAME, model_name)
    KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_MODEL_BASE_URL, base_url)
    if api_key:
        KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_MODEL_API_KEY, api_key)
    if format:
        KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_MODEL_FORMAT, format)
    KEY_FILE_HANDLER.write_key(KeyValues.USE_LOCAL_MODEL, "YES")
    KEY_FILE_HANDLER.write_key(KeyValues.USE_AZURE_OPENAI, "NO")
    print(
        ":raising_hands: Congratulations! You're now using a local model for all evals that require an LLM."
    )


@app.command(name="set-local-embeddings")
def set_local_embeddings_env(
    model_name: str = typer.Option(
        ..., "--model-name", help="Name of the local embeddings model"
    ),
    base_url: str = typer.Option(
        ..., "--base-url", help="Base URL for the local embeddings API"
    ),
    api_key: Optional[str] = typer.Option(
        None, "--api-key", help="API key for the local embeddings (if required)"
    ),
):
    KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_EMBEDDING_MODEL_NAME, model_name)
    KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_EMBEDDING_BASE_URL, base_url)
    if api_key:
        KEY_FILE_HANDLER.write_key(KeyValues.LOCAL_EMBEDDING_API_KEY, api_key)

    KEY_FILE_HANDLER.write_key(KeyValues.USE_LOCAL_EMBEDDINGS, "YES")
    KEY_FILE_HANDLER.write_key(KeyValues.USE_AZURE_OPENAI_EMBEDDING, "NO")

    print(
        ":raising_hands: Congratulations! You're now using local embeddings for all evals that require text embeddings."
    )


@app.command(name="unset-local-model")
def unset_local_model_env():
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_MODEL_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_MODEL_BASE_URL)
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_MODEL_API_KEY)
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_MODEL_FORMAT)
    KEY_FILE_HANDLER.remove_key(KeyValues.USE_LOCAL_MODEL)

    print(
        ":raising_hands: Congratulations! You're now using regular OpenAI for all evals that require an LLM."
    )


@app.command(name="unset-local-embeddings")
def unset_local_embeddings_env():
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_EMBEDDING_MODEL_NAME)
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_EMBEDDING_BASE_URL)
    KEY_FILE_HANDLER.remove_key(KeyValues.LOCAL_EMBEDDING_API_KEY)
    KEY_FILE_HANDLER.remove_key(KeyValues.USE_LOCAL_EMBEDDINGS)

    print(
        ":raising_hands: Congratulations! You're now using regular OpenAI embeddings for all evals that require text embeddings."
    )


if __name__ == "__main__":
    app()



from typing import Optional, Dict, Union, List
from deepeval.monitor.api import Link, CustomProperty, CustomPropertyType


def process_additional_data(
    additional_data: Optional[
        Dict[str, Union[str, Link, List[Link], Dict]]
    ] = None
):
    custom_properties = None
    if additional_data:
        custom_properties = {}
        for key, value in additional_data.items():
            if isinstance(value, str):
                custom_properties[key] = CustomProperty(
                    value=value, type=CustomPropertyType.TEXT
                )
            elif isinstance(value, dict):
                custom_properties[key] = CustomProperty(
                    value=value, type=CustomPropertyType.JSON
                )
            elif isinstance(value, Link):
                custom_properties[key] = CustomProperty(
                    value=value.value, type=CustomPropertyType.LINK
                )
            elif isinstance(value, list):
                if not all(isinstance(item, Link) for item in value):
                    raise ValueError(
                        "All values in 'additional_data' must be either of type 'string', 'Link', list of 'Link', or 'dict'."
                    )
                custom_properties[key] = [
                    CustomProperty(
                        value=item.value, type=CustomPropertyType.LINK
                    )
                    for item in value
                ]
            else:
                raise ValueError(
                    "All values in 'additional_data' must be either of type 'string', 'Link', list of 'Link', or 'dict'."
                )

    return custom_properties
